package com.example.session.repository;

import com.hazelcast.core.EntryEvent;
import com.hazelcast.core.HazelcastException;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.map.IMap;
import com.hazelcast.map.listener.EntryAddedListener;
import com.hazelcast.map.listener.EntryEvictedListener;
import com.hazelcast.map.listener.EntryRemovedListener;
import com.hazelcast.query.Predicates;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.session.FindByIndexNameSessionRepository;
import org.springframework.session.FlushMode;
import org.springframework.session.MapSession;
import org.springframework.session.Session;
import org.springframework.session.events.AbstractSessionEvent;
import org.springframework.session.events.SessionCreatedEvent;
import org.springframework.session.events.SessionDeletedEvent;
import org.springframework.session.events.SessionExpiredEvent;
import org.springframework.util.Assert;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.TimeUnit;

public class CustomHazelcastSessionRepository implements
        FindByIndexNameSessionRepository<CustomHazelcastSessionRepository.HazelcastSession>,
        EntryAddedListener<String, MapSession>, EntryEvictedListener<String, MapSession>,
        EntryRemovedListener<String, MapSession> {

    private static final Logger LOG = LoggerFactory.getLogger(CustomHazelcastSessionRepository.class);

    private boolean useEntryProcessor = false;

    public boolean isUseEntryProcessor() {
        return useEntryProcessor;
    }

    public void setUseEntryProcessor(boolean useEntryProcessor) {
        this.useEntryProcessor = useEntryProcessor;
    }

    /**
     * The default name of map used by Spring Session to store sessions.
     */
    public static final String DEFAULT_SESSION_MAP_NAME = "spring:session:sessions";

    /**
     * The principal name custom attribute name.
     */
    public static final String PRINCIPAL_NAME_ATTRIBUTE = "principalName";

    private static final Log logger = LogFactory.getLog(CustomHazelcastSessionRepository.class);

    private final HazelcastInstance hazelcastInstance;

    private ApplicationEventPublisher eventPublisher = new ApplicationEventPublisher() {

        @Override
        public void publishEvent(ApplicationEvent event) {
        }

        @Override
        public void publishEvent(Object event) {
        }

    };

    /**
     * If non-null, this value is used to override
     * {@link MapSession#setMaxInactiveInterval(Duration)}.
     */
    private Integer defaultMaxInactiveInterval;

    private String sessionMapName = DEFAULT_SESSION_MAP_NAME;

    private FlushMode flushMode = FlushMode.ON_SAVE;

    private IMap<String, MapSession> sessions;

    private UUID sessionListenerId;

    public CustomHazelcastSessionRepository(HazelcastInstance hazelcastInstance) {
        Assert.notNull(hazelcastInstance, "HazelcastInstance must not be null");
        this.hazelcastInstance = hazelcastInstance;
    }

    @PostConstruct
    public void init() {
        this.sessions = this.hazelcastInstance.getMap(this.sessionMapName);
        this.sessionListenerId = this.sessions.addEntryListener(this, true);
    }

    @PreDestroy
    public void close() {
        this.sessions.removeEntryListener(this.sessionListenerId);
    }

    /**
     * Sets the {@link ApplicationEventPublisher} that is used to publish
     * {@link AbstractSessionEvent session events}. The default is to not publish session
     * events.
     *
     * @param applicationEventPublisher the {@link ApplicationEventPublisher} that is used
     * to publish session events. Cannot be null.
     */
    public void setApplicationEventPublisher(
            ApplicationEventPublisher applicationEventPublisher) {
        Assert.notNull(applicationEventPublisher,
                "ApplicationEventPublisher cannot be null");
        this.eventPublisher = applicationEventPublisher;
    }

    /**
     * Set the maximum inactive interval in seconds between requests before newly created
     * sessions will be invalidated. A negative time indicates that the session will never
     * timeout. The default is 1800 (30 minutes).
     * @param defaultMaxInactiveInterval the maximum inactive interval in seconds
     */
    public void setDefaultMaxInactiveInterval(Integer defaultMaxInactiveInterval) {
        this.defaultMaxInactiveInterval = defaultMaxInactiveInterval;
    }

    /**
     * Set the name of map used to store sessions.
     * @param sessionMapName the session map name
     */
    public void setSessionMapName(String sessionMapName) {
        Assert.hasText(sessionMapName, "Map name must not be empty");
        this.sessionMapName = sessionMapName;
    }

    /**
     * Sets the Hazelcast flush mode. Default flush mode is {@link FlushMode#ON_SAVE}.
     * @param flushMode the new Hazelcast flush mode
     */
    public void setFlushMode(FlushMode flushMode) {
        Assert.notNull(flushMode, "flushMode cannot be null");
        this.flushMode = flushMode;
    }

    @Override
    public CustomHazelcastSessionRepository.HazelcastSession createSession() {
        CustomHazelcastSessionRepository.HazelcastSession result = new CustomHazelcastSessionRepository.HazelcastSession();
        if (this.defaultMaxInactiveInterval != null) {
            result.setMaxInactiveInterval(
                    Duration.ofSeconds(this.defaultMaxInactiveInterval));
        }
        return result;
    }

    @Override
    public void save(CustomHazelcastSessionRepository.HazelcastSession session) {

        try {

            if (session.isNew) {
                this.sessions.set(session.getId(), session.getDelegate(),
                        session.getMaxInactiveInterval().getSeconds(), TimeUnit.SECONDS);
            } else if (session.sessionIdChanged) {
                this.sessions.delete(session.originalId);
                session.originalId = session.getId();
                this.sessions.set(session.getId(), session.getDelegate(),
                        session.getMaxInactiveInterval().getSeconds(), TimeUnit.SECONDS);
            } else if (session.hasChanges()) {

                if(this.useEntryProcessor) {
                    // entity processor is executed on the host where the data is located
                    CustomSessionUpdateEntryProcessor entryProcessor = new CustomSessionUpdateEntryProcessor();
                    if (session.lastAccessedTimeChanged) {
                        entryProcessor.setLastAccessedTime(session.getLastAccessedTime());
                    }
                    if (session.maxInactiveIntervalChanged) {
                        entryProcessor.setMaxInactiveInterval(session.getMaxInactiveInterval());
                    }
                    if (!session.delta.isEmpty()) {
                        entryProcessor.setDelta(session.delta);
                    }
                    this.sessions.executeOnKey(session.getId(), entryProcessor);
                } else {
                    // direct put operation ( is not atomic )
                    this.sessions.put(session.getId(), session.getDelegate(), session.getMaxInactiveInterval().getSeconds(), TimeUnit.SECONDS);
                }
            }
            session.clearChangeFlags();

        } catch(HazelcastException | IllegalStateException e){
            LOG.error("hazelcast session failed to save", e);
        }
    }

    @Override
    public CustomHazelcastSessionRepository.HazelcastSession findById(String id) {

        try {
            MapSession saved = this.sessions.get(id);
            if (saved == null) {
                return null;
            }
            if (saved.isExpired()) {
                deleteById(saved.getId());
                return null;
            }
            return new CustomHazelcastSessionRepository.HazelcastSession(saved);
        } catch(HazelcastException | IllegalStateException e){
            LOG.error("hazelcast session failed to find", e);

            return null;
        }
    }

    @Override
    public void deleteById(String id) {
        try {
            this.sessions.remove(id);
        } catch(HazelcastException | IllegalStateException e){
            LOG.error("hazelcast session failed to delete", e);
        }
    }

    @Override
    public Map<String, CustomHazelcastSessionRepository.HazelcastSession> findByIndexNameAndIndexValue(String indexName, String indexValue) {
        try {
            if (!PRINCIPAL_NAME_INDEX_NAME.equals(indexName)) {
                return Collections.emptyMap();
            }
            Collection<MapSession> sessions = this.sessions
                    .values(Predicates.equal(PRINCIPAL_NAME_ATTRIBUTE, indexValue));
            Map<String, CustomHazelcastSessionRepository.HazelcastSession> sessionMap = new HashMap<>(sessions.size());
            for (MapSession session : sessions) {
                sessionMap.put(session.getId(), new CustomHazelcastSessionRepository.HazelcastSession(session));
            }
            return sessionMap;
        } catch(HazelcastException | IllegalStateException e){
            LOG.error("hazelcast session failed findByIndexNameAndIndexValue", e);
            return new HashMap<>();
        }
    }

    @Override
    public void entryAdded(EntryEvent<String, MapSession> event) {
        MapSession session = event.getValue();
        if (session.getId().equals(session.getOriginalId())) {
            if (logger.isDebugEnabled()) {
                logger.debug("Session created with id: " + session.getId());
            }
            this.eventPublisher.publishEvent(new SessionCreatedEvent(this, session));
        }
    }

    @Override
    public void entryEvicted(EntryEvent<String, MapSession> event) {
        if (logger.isDebugEnabled()) {
            logger.debug("Session expired with id: " + event.getOldValue().getId());
        }
        this.eventPublisher
                .publishEvent(new SessionExpiredEvent(this, event.getOldValue()));
    }

    @Override
    public void entryRemoved(EntryEvent<String, MapSession> event) {
        MapSession session = event.getOldValue();
        if (session != null) {
            if (logger.isDebugEnabled()) {
                logger.debug("Session deleted with id: " + session.getId());
            }
            this.eventPublisher.publishEvent(new SessionDeletedEvent(this, session));
        }
    }

    /**
     * A custom implementation of {@link Session} that uses a {@link MapSession} as the
     * basis for its mapping. It keeps track if changes have been made since last save.
     *
     * @author Aleksandar Stojsavljevic
     */
    final class HazelcastSession implements Session {

        private final MapSession delegate;

        private boolean isNew;

        private boolean sessionIdChanged;

        private boolean lastAccessedTimeChanged;

        private boolean maxInactiveIntervalChanged;

        private String originalId;

        private Map<String, Object> delta = new HashMap<>();

        /**
         * Creates a new instance ensuring to mark all of the new attributes to be
         * persisted in the next save operation.
         */
        HazelcastSession() {
            this(new MapSession());
            this.isNew = true;
            flushImmediateIfNecessary();
        }

        /**
         * Creates a new instance from the provided {@link MapSession}.
         * @param cached the {@link MapSession} that represents the persisted session that
         * was retrieved. Cannot be {@code null}.
         */
        HazelcastSession(MapSession cached) {
            Assert.notNull(cached, "MapSession cannot be null");
            this.delegate = cached;
            this.originalId = cached.getId();
        }

        @Override
        public void setLastAccessedTime(Instant lastAccessedTime) {
            this.delegate.setLastAccessedTime(lastAccessedTime);
            this.lastAccessedTimeChanged = true;
            flushImmediateIfNecessary();
        }

        @Override
        public boolean isExpired() {
            return this.delegate.isExpired();
        }

        @Override
        public Instant getCreationTime() {
            return this.delegate.getCreationTime();
        }

        @Override
        public String getId() {
            return this.delegate.getId();
        }

        @Override
        public String changeSessionId() {
            String newSessionId = this.delegate.changeSessionId();
            this.sessionIdChanged = true;
            return newSessionId;
        }

        @Override
        public Instant getLastAccessedTime() {
            return this.delegate.getLastAccessedTime();
        }

        @Override
        public void setMaxInactiveInterval(Duration interval) {
            this.delegate.setMaxInactiveInterval(interval);
            this.maxInactiveIntervalChanged = true;
            flushImmediateIfNecessary();
        }

        @Override
        public Duration getMaxInactiveInterval() {
            return this.delegate.getMaxInactiveInterval();
        }

        @Override
        public <T> T getAttribute(String attributeName) {
            return this.delegate.getAttribute(attributeName);
        }

        @Override
        public Set<String> getAttributeNames() {
            return this.delegate.getAttributeNames();
        }

        @Override
        public void setAttribute(String attributeName, Object attributeValue) {
            this.delegate.setAttribute(attributeName, attributeValue);
            this.delta.put(attributeName, attributeValue);
            flushImmediateIfNecessary();
        }

        @Override
        public void removeAttribute(String attributeName) {
            this.delegate.removeAttribute(attributeName);
            this.delta.put(attributeName, null);
            flushImmediateIfNecessary();
        }

        MapSession getDelegate() {
            return this.delegate;
        }

        boolean hasChanges() {
            return (this.lastAccessedTimeChanged || this.maxInactiveIntervalChanged
                    || !this.delta.isEmpty());
        }

        void clearChangeFlags() {
            this.isNew = false;
            this.lastAccessedTimeChanged = false;
            this.sessionIdChanged = false;
            this.maxInactiveIntervalChanged = false;
            this.delta.clear();
        }

        private void flushImmediateIfNecessary() {
            if (CustomHazelcastSessionRepository.this.flushMode == FlushMode.IMMEDIATE) {
                CustomHazelcastSessionRepository.this.save(this);
            }
        }

    }

}
